/*
 *   Display.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
// Display driver. Update is done in real-time at each acces in video-memory.
// CRT refresh frequency is emulated with a a timer based on Capricorn cycle
// counter on real-speed and with High performance counter otherwise at 1/60 sec.
//
// HP87
// Zoom 1 and 2 classic display, new mode with VRAM64K in zoom 2
//
// scan time 10217 with 827 cycles as redraw
//
// HP85
//
// Zoom 1, 2 and 3 classic display
//
// scan time 36*(240-0-46) cycle with 36*(70-0-15) cycles as redraw (from service rom)
//

#include "StdAfx.h"
#include "HP8x.h"
#include "resource.h"
#include "io.h"
#include "kml.h"

// #define DEBUG_DISPLAY					// switch for DISPLAY debug purpose

#define NOCOLORS	2

#define B 0x00000000						// black
#define W 0x00FFFFFF						// white
#define I 0xFFFFFFFF						// ignore

static HBITMAP hLcdBitmap;					// screen bitmap
static HBITMAP hMainBitmap;					// main bitmap
static HBITMAP hFontBitmap;					// font Bitmap
static HBITMAP hOldMainBitmap;

static DWORD dwKMLColor[64] =				// color table loaded by KML script
{
	B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,
	W,W,W,W,W,W,W,W,W,W,W,W,W,W,W,W,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I
};

static struct
{
	BITMAPINFOHEADER Lcd_bmih;
	RGBQUAD bmiColors[NOCOLORS]; 
} bmiLcd =
{
	{0x28,0/*x*/,0/*y*/,1,8,BI_RGB,0,0,0,NOCOLORS,0}
};

static  LPBYTE	pbyLcd;						// address of bitmap for the screen
static  HDC		hFontDC = NULL;				// handle of display font

// size of display at Zoom 0, 1, 2, 3

#if defined HP87
INT nLcdCol[4] = {640, 640, 1088,1016};
INT nLcdRow[4] = {240, 240,  480, 480};
#endif
#if defined HP85
INT nLcdCol[4] = {256, 256, 512, 768};
INT nLcdRow[4] = {192, 192, 384, 576};
#endif

UINT    nBackgroundX = 0;					// offset of background X (from KML script)
UINT    nBackgroundY = 0;					// offset of background Y (from KML script)
UINT    nBackgroundW = 0;					// width of background (from KML script)
UINT    nBackgroundH = 0;					// height of background (from KML script)
UINT    nLcdX = 0;							// offset of screen X (from KML script)
UINT    nLcdY = 0;							// offset of screen Y (from KML script)
UINT    nLcdxZoom = 1;						// zoom on X (from KML script)
UINT	nLcdyZoom = 1;						// zoom on Y (from KML script)
HDC     hLcdDC = NULL;						// handle of bitmap for screen
HDC     hMainDC = NULL;						// handle of main bitmap for window display

BOOL	bOffScreen = FALSE;

//################
//#
//#    Low level subroutines
//#
//################

//
// setting of contrast (not fully used)
//
static VOID UpdateContrast(BYTE byContrast)
{
	bmiLcd.bmiColors[1] = *(RGBQUAD*)&dwKMLColor[0];    // pixel on  color
	bmiLcd.bmiColors[0] = *(RGBQUAD*)&dwKMLColor[32]; // pixel off color
	// update palette information
	_ASSERT(hLcdDC);
	SetDIBColorTable(hLcdDC,0,ARRAYSIZEOF(bmiLcd.bmiColors),bmiLcd.bmiColors); 
}

//################
//#
//#    Public functions
//#
//################

//
// settings for palette colors
//
VOID SetLcdColor(UINT nId, UINT nRed, UINT nGreen, UINT nBlue)
{
	dwKMLColor[nId&0x3F] = ((nRed&0xFF)<<16)|((nGreen&0xFF)<<8)|(nBlue&0xFF);
}

//
// allocate screen bitmap and font
//
VOID CreateLcdBitmap(VOID)
{
	// create screen bitmap
	bmiLcd.Lcd_bmih.biWidth = nLcdCol[nLcdxZoom];
	bmiLcd.Lcd_bmih.biHeight = -nLcdRow[nLcdyZoom];
	_ASSERT(hLcdDC == NULL);
	hLcdDC = CreateCompatibleDC(hWindowDC);
	_ASSERT(hLcdDC != NULL);
	hLcdBitmap = CreateDIBSection(hLcdDC, (BITMAPINFO*)&bmiLcd,DIB_RGB_COLORS, (LPVOID*)&pbyLcd, NULL, 0);
	_ASSERT(hLcdBitmap != NULL);
	hLcdBitmap = SelectObject(hLcdDC, hLcdBitmap);
	_ASSERT(hPalette != NULL);
	SelectPalette(hLcdDC, hPalette, FALSE);	// set palette for screen DC
	RealizePalette(hLcdDC);					// realize palette
	hFontDC = CreateCompatibleDC(hWindowDC);
	_ASSERT(hFontDC != NULL);

#if defined HP87
	if (nLcdxZoom == 1)
		hFontBitmap = LoadImage(hApp,MAKEINTRESOURCE(IDB_87_FONTMAP1), IMAGE_BITMAP, 0, 0, LR_MONOCHROME);	// Load bitmap font
	else
#if defined VRAM64K
		hFontBitmap = LoadImage(hApp,MAKEINTRESOURCE(IDB_87_FONTMAP3), IMAGE_BITMAP, 0, 0, LR_MONOCHROME);	// Load bitmap font		
#else
		hFontBitmap = LoadImage(hApp,MAKEINTRESOURCE(IDB_87_FONTMAP2), IMAGE_BITMAP, 0, 0, LR_MONOCHROME);	// Load bitmap font		
#endif
#endif

#if defined HP85
	if (nLcdxZoom == 1)
		hFontBitmap = LoadImage(hApp,MAKEINTRESOURCE(IDB_85_FONTMAP1), IMAGE_BITMAP, 0, 0, LR_MONOCHROME);	// Load bitmap font
	else if (nLcdxZoom == 2)
		hFontBitmap = LoadImage(hApp,MAKEINTRESOURCE(IDB_85_FONTMAP2), IMAGE_BITMAP, 0, 0, LR_MONOCHROME);	// Load bitmap font		
	else if (nLcdxZoom == 3)
		hFontBitmap = LoadImage(hApp,MAKEINTRESOURCE(IDB_85_FONTMAP3), IMAGE_BITMAP, 0, 0, LR_MONOCHROME);	// Load bitmap font		
#endif
	_ASSERT(hFontBitmap != NULL);
	hFontBitmap = SelectObject(hFontDC, hFontBitmap);
	UpdateContrast(23);
}

//
// free screen bitmap and font
//
VOID DestroyLcdBitmap(VOID)
{
	// set contrast palette to startup colors
	WORD i = 0;   dwKMLColor[i++] = W;
	while(i < 32) dwKMLColor[i++] = B;
	while(i < 64) dwKMLColor[i++] = I;

	if (hLcdDC != NULL)
	{
		// destroy screen bitmap
		SelectObject(hLcdDC, hLcdBitmap);
		DeleteObject(hLcdBitmap);
		DeleteDC(hLcdDC);
		hLcdDC = NULL;
		hLcdBitmap = NULL;
	}
	if (hFontDC != NULL)
	{
		// destroy font bitmap
		SelectObject(hFontDC, hFontBitmap);
		DeleteObject(hFontBitmap);
		DeleteDC(hFontDC);
		hFontDC = NULL;
		hFontBitmap = NULL;
	}
}

//
// allocate main bitmap
//
BOOL CreateMainBitmap(LPCTSTR szFilename)
{
	HPALETTE hAssertPalette;

	_ASSERT(hWindowDC != NULL);
	hMainDC = CreateCompatibleDC(hWindowDC);
	_ASSERT(hMainDC != NULL);
	if (hMainDC == NULL) return FALSE;		// quit if failed
	hMainBitmap = LoadBitmapFile(szFilename);
	if (hMainBitmap == NULL)
	{
		DeleteDC(hMainDC);
		return FALSE;
	}
	hMainBitmap = SelectObject(hMainDC, hMainBitmap);
	_ASSERT(hPalette != NULL);
	hAssertPalette = SelectPalette(hMainDC, hPalette, FALSE);
	_ASSERT(hAssertPalette != NULL);
	RealizePalette(hMainDC);
	return TRUE;
}

//
// free main bitmap
//
VOID DestroyMainBitmap(VOID)
{
	if (hMainDC != NULL)
	{
		// destroy Main bitmap
		SelectObject(hMainDC, hMainBitmap);
		DeleteObject(hMainBitmap);
		DeleteDC(hMainDC);
		hMainDC = NULL;
		hMainBitmap = NULL;
		hOldMainBitmap = NULL;
	}
}

//
// update display (bForce TRUE: only a section FALSE: full screen update)
//
VOID UpdateMainDisplay(BOOL bForce)
{
	WORD i, j, d, n, n_lines;
	INT   x0;			
	INT   y0;
	INT   hc;					// height of alpha line (10, 15, 20, 30 pixels) 
								// or horizontal offset of the beginning of a graphic line
	LPBYTE pLcd, pVid;
	DWORD mask;					// for wrap on video memory

	#if defined DEBUG_DISPLAY
	{
		TCHAR buffer[256];
		wsprintf(buffer,_T("%04X: Update Main Display, %d\n"), Chipset.R[4]|(Chipset.R[5]<<8), force);
		OutputDebugString(buffer);
	}
	#endif


#if defined HP87
	if ((Chipset.crtsts & 0x06) != 0x00)					// crt blank (no beam) or power
	{
		if (!bOffScreen)
			ZeroMemory(pbyLcd, nLcdCol[nLcdxZoom] * nLcdRow[nLcdyZoom]);	// clear screen bitmap
		bOffScreen = TRUE;
		goto doit;
	}
#endif
#if defined HP85
	if ((Chipset.crtwsts & 0x06) != 0x00)					// crt wipe (no beam) or power
	{
		if (bRealSpeed)
		{
			if (!bOffScreen)
				ZeroMemory(pbyLcd, nLcdCol[nLcdxZoom] * nLcdRow[nLcdyZoom]);	// clear screen bitmap
			bOffScreen = TRUE;
			goto doit; 
		}
		else
			goto nothing;
	}
#endif
	else													// update something
	{
		bOffScreen = FALSE;

#if defined HP87
		if ((Chipset.crtsts & 0x80) == 0x00)					// alpha mode ?
		{
			if (nLcdxZoom == 1)										// zoom 1
				hc = ((Chipset.crtsts & 0x08) == 0x08) ? 10:15;			// height of character	(pagesize 24 or 16)
			else													// zoom 2
				hc = ((Chipset.crtsts & 0x08) == 0x08) ? 20:30;			// height of character	(pagesize 24 or 16)
			mask = ((Chipset.crtsts & 0x40) == 0x40) ? 0x3FC0:0x10E0;	// 54 or 204 lines		(alpha or alphall)

			n_lines = ((Chipset.crtsts & 0x08) == 0x08) ? 24:16;	// 24 or 16 lines to redraw
			pVid = Chipset.Video + (DWORD) (Chipset.crtsad);		// address of beginning of screen update
			y0 = 0;													
			x0 = 0;													// character of start of update (always 0)

			if (!bForce)											// just a change on CRTSAD
			{
				if (Chipset.crtsad > Chipset.crtoldsad)					// display scroll up ?	
				{
					d = Chipset.crtsad - Chipset.crtoldsad;					
					if ((d % 80) == 0)										// full line move ?
					{
						if ((d / 80) < n_lines)									// less than 24 or 16 lines ?
						{
							n = n_lines - (d / 80);									// number of lines to move
							n_lines -= n;											// number of lines to update
							pVid += 80 * n;											// beginning of draw
							BitBlt(hLcdDC, 0, y0, nLcdCol[nLcdxZoom], hc*n, hLcdDC, 0, y0+hc*n_lines, SRCCOPY);	// move 
							y0 += n*hc;												// line of start of update
						}
					}
				}
				else													// display scroll down ?
				{
					d = Chipset.crtoldsad - Chipset.crtsad;
					if ((d % 80) == 0)
					{
						if ((d / 80) < n_lines)
						{
							n = n_lines - (d / 80);									// number of lines to move
							n_lines -= n;											// number of lines to update
							BitBlt(hLcdDC, 0, y0+hc*n_lines, nLcdCol[nLcdxZoom], hc*n, hLcdDC, 0, y0, SRCCOPY);
						}
					}
				}
			}
			Chipset.crtoldsad = Chipset.crtsad;						// update old screen start address in alpha mode
			n_lines *= 80;											// number of characters to update
			if (pVid >= Chipset.Video + mask) pVid -= mask;			// wrap ?
			if ((Chipset.crtsts & 0x20) == 0x20)					// inverse video mode ?
			{
				if (nLcdxZoom == 1)										// zoom 1
				{
					for (i = 0; i < n_lines; i++)							// for all characters
					{
						BitBlt(hLcdDC, x0, y0, 8, 10, hFontDC, (*pVid++) << 3, 0, NOTSRCCOPY);	// copy a bitmap character
						x0 += 8;												// next character
						if (x0 == 640)											// end of line ?
						{
							y0 += hc;												// next charcater line
							x0 = 0;													// first character of the new line
						}
						if (pVid >= Chipset.Video + mask) pVid = Chipset.Video;	// wrap ?
					}
				}
				else													// zoom 2 or 3
				{
					for (i = 0; i < n_lines; i++)
					{
#if defined VRAM64K
						BitBlt(hLcdDC, 28+x0, y0, 12, 20, hFontDC, ((*pVid++) * 12), 0, NOTSRCCOPY);
						x0 += 12;
						if (x0 == 960)
#else
						BitBlt(hLcdDC, 24+x0, y0, 13, 20, hFontDC, ((*pVid++) << 4) + 3, 0, NOTSRCCOPY);
						x0 += 13;
						if (x0 == 1040)
#endif
						{
							y0 += hc;
							x0 = 0;
						}
						if (pVid >= Chipset.Video + mask) pVid = Chipset.Video;
					}
				}
			}
			else													// normal video mode
			{
				if (nLcdxZoom == 1)
				{
					for (i = 0; i < n_lines; i++)
					{
						BitBlt(hLcdDC, x0, y0, 8, 10, hFontDC, (*pVid++) << 3, 0, SRCCOPY);
						x0 += 8;
						if (x0 == 640)
						{
							y0 += hc;
							x0 = 0;
						}
						if (pVid >= Chipset.Video + mask) pVid = Chipset.Video;
					}
				}
				else
				{
					for (i = 0; i < n_lines; i++)
					{
#if defined VRAM64K
						BitBlt(hLcdDC, 28+x0, y0, 12, 20, hFontDC, ((*pVid++) * 12), 0, SRCCOPY);
						x0 += 12;
						if (x0 == 960)
#else
						BitBlt(hLcdDC, 24+x0, y0, 13, 20, hFontDC, ((*pVid++) << 4) + 3, 0, SRCCOPY);
						x0 += 13;
						if (x0 == 1040)
#endif
						{
							y0 += hc;
							x0 = 0;
						}
						if (pVid >= Chipset.Video + mask) pVid = Chipset.Video;
					}
				}
			}
		}
		else													// graph mode 
		{
			pVid = Chipset.Video + 0x10E0;							// start address
			if ((Chipset.crtsts & 0x40) == 0x40)					// graphall mode ?
			{
				if (nLcdxZoom == 1)										// zoom 1
					hc = 96;												// horizontal offset (640 - 544)
				else													// zoom 2 or 3
#if defined VRAM64K
					hc = 0;
				n_lines = 127;
#else
					hc = 0;													// horizontal ofset (1088 - 1088)
				n_lines = 68;												// 68 bytes per line (8*68 = 544)
#endif
			}
			else													// graph mode
			{
				if (nLcdxZoom == 1)										// zoom 1
					hc = 240;												// horizontal offset (640 - 400)
				else													// zoom 2 or 3
#if defined VRAM64K
					hc = 0;
				n_lines = 127;
#else
					hc = 288;												// horizontal offset (1088 - 800)
				n_lines = 50;												// 50 bytes per line (8*50 = 400)
#endif
			}
			pLcd = pbyLcd + hc/2;									// screean bitmap start
			if ((Chipset.crtsts & 0x20) == 0x20)					// inverse video mode ?
			{
				if (nLcdxZoom == 1)										// zoom 1
				{		
					for (i = 0; i < 240; i++)								// 240 line to display
					{
						for (j= 0; j < n_lines; j++)							// n_lines bytes per line
						{
							*pLcd++ = 1 - ((*pVid & 0x80)>>7);						// first pixel
							*pLcd++ = 1 - ((*pVid & 0x40)>>6);						// second pixel
							*pLcd++ = 1 - ((*pVid & 0x20)>>5);						// ...
							*pLcd++ = 1 - ((*pVid & 0x10)>>4);
							*pLcd++ = 1 - ((*pVid & 0x08)>>3);
							*pLcd++ = 1 - ((*pVid & 0x04)>>2);
							*pLcd++ = 1 - ((*pVid & 0x02)>>1);
							*pLcd++ = 1 - (*pVid & 0x01);
							pVid++;
							if (pVid >= Chipset.Video + 0x4000) pVid=Chipset.Video;	// wrap ?
						}
						pLcd += hc;												// offset for net screen line
					}
				}
				else													// zoom 2 or 3
				{
					LPBYTE mem;												// for screen bitmap

#if defined VRAM64K
					for (i = 0; i < 480; i++)								// 480 lines to update
#else
					for (i = 0; i < 240; i++)								// 240 lines to update
#endif
					{
						mem = pVid;												// screen address start of line	
						for (j= 0; j < n_lines; j++)							// n_lines bytes per line
						{	
							*pLcd++ = 1 - ((*pVid & 0x80)>>7);						// first pixel first time
#if !defined VRAM64K
							*pLcd++ = 1 - ((*pVid & 0x80)>>7);						// first pixel second times
#endif
							*pLcd++ = 1 - ((*pVid & 0x40)>>6);						// second pixel first time
#if !defined VRAM64K
							*pLcd++ = 1 - ((*pVid & 0x40)>>6);						// ...
#endif
							*pLcd++ = 1 - ((*pVid & 0x20)>>5);
#if !defined VRAM64K
							*pLcd++ = 1 - ((*pVid & 0x20)>>5);
#endif
							*pLcd++ = 1 - ((*pVid & 0x10)>>4);
#if !defined VRAM64K
							*pLcd++ = 1 - ((*pVid & 0x10)>>4);
#endif
							*pLcd++ = 1 - ((*pVid & 0x08)>>3);
#if !defined VRAM64K
							*pLcd++ = 1 - ((*pVid & 0x08)>>3);
#endif
							*pLcd++ = 1 - ((*pVid & 0x04)>>2);
#if !defined VRAM64K
							*pLcd++ = 1 - ((*pVid & 0x04)>>2);
#endif
							*pLcd++ = 1 - ((*pVid & 0x02)>>1);
#if !defined VRAM64K
							*pLcd++ = 1 - ((*pVid & 0x02)>>1);
#endif
							*pLcd++ = 1 - (*pVid & 0x01);
#if !defined VRAM64K
							*pLcd++ = 1 - (*pVid & 0x01);
#endif
							pVid++;
#if !defined VRAM64K
							if (pVid >= Chipset.Video + 0x4000) pVid=Chipset.Video;	// wrap ?
#endif
						}
						pLcd += hc;												// next screen line
#if !defined VRAM64K
						pVid = mem;												// redo the same line a second time
						for (j= 0; j < n_lines; j++)							// n_lines bytes per line
						{
							*pLcd++ = 1 - ((*pVid & 0x80)>>7);						// first pixel firts time
							*pLcd++ = 1 - ((*pVid & 0x80)>>7);						// first pixel second time
							*pLcd++ = 1 - ((*pVid & 0x40)>>6);						// second pixel first time
							*pLcd++ = 1 - ((*pVid & 0x40)>>6);						// ...
							*pLcd++ = 1 - ((*pVid & 0x20)>>5);
							*pLcd++ = 1 - ((*pVid & 0x20)>>5);
							*pLcd++ = 1 - ((*pVid & 0x10)>>4);
							*pLcd++ = 1 - ((*pVid & 0x10)>>4);
							*pLcd++ = 1 - ((*pVid & 0x08)>>3);
							*pLcd++ = 1 - ((*pVid & 0x08)>>3);
							*pLcd++ = 1 - ((*pVid & 0x04)>>2);
							*pLcd++ = 1 - ((*pVid & 0x04)>>2);
							*pLcd++ = 1 - ((*pVid & 0x02)>>1);
							*pLcd++ = 1 - ((*pVid & 0x02)>>1);
							*pLcd++ = 1 - (*pVid & 0x01);
							*pLcd++ = 1 - (*pVid & 0x01);
							pVid++;
							if (pVid >= Chipset.Video + 0x4000) pVid=Chipset.Video;	// wrap ?
						}
						pLcd += hc;												// next screen line
#endif
					}
				}
			}
			else													// normal video mode
			{
				if (nLcdxZoom == 1)
				{
					for (i = 0; i < 240; i++)
					{
						for (j= 0; j < n_lines; j++)
						{
							*pLcd++ = (*pVid & 0x80)>>7;
							*pLcd++ = (*pVid & 0x40)>>6;
							*pLcd++ = (*pVid & 0x20)>>5;
							*pLcd++ = (*pVid & 0x10)>>4;
							*pLcd++ = (*pVid & 0x08)>>3;
							*pLcd++ = (*pVid & 0x04)>>2;
							*pLcd++ = (*pVid & 0x02)>>1;
							*pLcd++ = (*pVid & 0x01);
							pVid++;
							if (pVid >= Chipset.Video + 0x4000) pVid=Chipset.Video;
						}
						pLcd += hc;
					}
				}
				else
				{
					LPBYTE mem;

#if defined VRAM64K
					for (i = 0; i < 480; i++)								// 480 lines to update
#else
					for (i = 0; i < 240; i++)								// 240 lines to update
#endif
					{
						mem = pVid;
						for (j= 0; j < n_lines; j++)
						{
							*pLcd++ = (*pVid & 0x80)>>7;
#if !defined VRAM64K
							*pLcd++ = (*pVid & 0x80)>>7;
#endif
							*pLcd++ = (*pVid & 0x40)>>6;
#if !defined VRAM64K
							*pLcd++ = (*pVid & 0x40)>>6;
#endif
							*pLcd++ = (*pVid & 0x20)>>5;
#if !defined VRAM64K
							*pLcd++ = (*pVid & 0x20)>>5;
#endif
							*pLcd++ = (*pVid & 0x10)>>4;
#if !defined VRAM64K
							*pLcd++ = (*pVid & 0x10)>>4;
#endif
							*pLcd++ = (*pVid & 0x08)>>3;
#if !defined VRAM64K
							*pLcd++ = (*pVid & 0x08)>>3;
#endif
							*pLcd++ = (*pVid & 0x04)>>2;
#if !defined VRAM64K
							*pLcd++ = (*pVid & 0x04)>>2;
#endif
							*pLcd++ = (*pVid & 0x02)>>1;
#if !defined VRAM64K
							*pLcd++ = (*pVid & 0x02)>>1;
#endif
							*pLcd++ = (*pVid & 0x01);
#if !defined VRAM64K
							*pLcd++ = (*pVid & 0x01);
#endif
							pVid++;
#if !defined VRAM64K
							if (pVid >= Chipset.Video + 0x4000) pVid=Chipset.Video;
#endif
						}
						pLcd += hc;
#if !defined VRAM64K
						pVid = mem;
						for (j= 0; j < n_lines; j++)
						{
							*pLcd++ = (*pVid & 0x80)>>7;
							*pLcd++ = (*pVid & 0x80)>>7;
							*pLcd++ = (*pVid & 0x40)>>6;
							*pLcd++ = (*pVid & 0x40)>>6;
							*pLcd++ = (*pVid & 0x20)>>5;
							*pLcd++ = (*pVid & 0x20)>>5;
							*pLcd++ = (*pVid & 0x10)>>4;
							*pLcd++ = (*pVid & 0x10)>>4;
							*pLcd++ = (*pVid & 0x08)>>3;
							*pLcd++ = (*pVid & 0x08)>>3;
							*pLcd++ = (*pVid & 0x04)>>2;
							*pLcd++ = (*pVid & 0x04)>>2;
							*pLcd++ = (*pVid & 0x02)>>1;
							*pLcd++ = (*pVid & 0x02)>>1;
							*pLcd++ = (*pVid & 0x01);
							*pLcd++ = (*pVid & 0x01);
							pVid++;
							if (pVid >= Chipset.Video + 0x4000) pVid=Chipset.Video;
						}
						pLcd += hc;
#endif
					}
				}
			}
		}
	}
#endif

#if defined HP85
		if ((Chipset.crtwsts & 0x80) == 0x00)		// alpha mode	
		{
			if (nLcdxZoom == 1)
				hc = 12;
			else if (nLcdxZoom == 2)
				hc = 24;
			else if (nLcdxZoom == 3)
				hc = 36;
			mask = 0x0800;							// 64 lines

			n_lines = 16;							// 16 line / screen
			pVid = Chipset.Video + (DWORD) (Chipset.crtsad >> 1);											// beginning of screen
			y0 = 0;
			x0 = 0;

			if (!bForce)	// just a change on CRTSAD
			{
				if ((Chipset.crtsad >> 1) > (Chipset.crtoldsad >> 1))
				{
					d = (Chipset.crtsad >> 1) - (Chipset.crtoldsad >> 1);
					if ((d % 32) == 0)
					{
						if ((d / 32) < n_lines)
						{
							n = n_lines - (d / 32);			// number of lines to move
							n_lines -= n;					// number of lines to draw
							pVid += 32 * n;					// beginning of draw
							BitBlt(hLcdDC, 0, y0, nLcdCol[nLcdxZoom], hc*n, hLcdDC, 0, y0+hc*n_lines, SRCCOPY);
							y0 += n*hc;
						}
					}
				}
				else
				{
					d = (Chipset.crtoldsad >> 1) - (Chipset.crtsad >> 1);
					if ((d % 32) == 0)
					{
						if ((d / 32) < n_lines)
						{
							n = n_lines - (d / 32);			// number of lines to move
							n_lines -= n;					// number of lines to draw
							BitBlt(hLcdDC, 0, y0+hc*n_lines, nLcdCol[nLcdxZoom], hc*n, hLcdDC, 0, y0, SRCCOPY);
						}
					}
				}
			}
			Chipset.crtoldsad = Chipset.crtsad;

			// scan new display area
			n_lines *= 32;
			if (pVid >= Chipset.Video + mask) pVid -= mask;
			if (nLcdxZoom == 1)
			{
				for (i = 0; i < n_lines; i++)
				{
					BitBlt(hLcdDC, x0, y0, 8, 12, hFontDC, (*pVid++) << 3, 0, SRCCOPY);
					x0 += 8;
					if (x0 == 256)
					{
						y0 += hc;
						x0 = 0;
					}
					if (pVid >= Chipset.Video + mask) pVid -= mask;
				}
			}
			else if (nLcdxZoom == 2)
			{
				for (i = 0; i < n_lines; i++)
				{
					BitBlt(hLcdDC, x0, y0, 16, 24, hFontDC, (*pVid++) * 16, 0, SRCCOPY);
					x0 += 16;
					if (x0 == 512)
					{
						y0 += hc;
						x0 = 0;
					}
					if (pVid >= Chipset.Video + mask) pVid -= mask;
				}
			}
			else if (nLcdxZoom == 3)
			{
				for (i = 0; i < n_lines; i++)
				{
					BitBlt(hLcdDC, x0, y0, 24, 36, hFontDC, (*pVid++) * 24, 0, SRCCOPY);
					x0 += 24;
					if (x0 == 768)
					{
						y0 += hc;
						x0 = 0;
					}
					if (pVid >= Chipset.Video + mask) pVid -= mask;
				}
			}
		}
		else
		{
			pVid = Chipset.Video + 0x0800;
			hc = 0;
			n_lines = 32;
			pLcd = pbyLcd;
			if (nLcdxZoom == 1)
			{
				for (i = 0; i < 192; i++)
				{
					for (j= 0; j < n_lines; j++)
					{
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x01);
						pVid++;
						if (pVid >= Chipset.Video + 0x2000) pVid=Chipset.Video + 0x0800;
					}
					pLcd += hc;
				}
			}
			else if (nLcdxZoom == 2)
			{
				LPBYTE mem;

				for (i = 0; i < 192; i++)
				{
					mem = pVid;
					for (j= 0; j < n_lines; j++)
					{
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01);
						pVid++;
						if (pVid >= Chipset.Video + 0x2000) pVid=Chipset.Video + 0x0800;
					}
					pLcd += hc;
					pVid = mem;
					for (j= 0; j < n_lines; j++)
					{
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01);
						pVid++;
						if (pVid >= Chipset.Video + 0x2000) pVid=Chipset.Video + 0x0800;
					}
					pLcd += hc;
				}
			}
			else if (nLcdxZoom == 3)
			{
				LPBYTE mem;

				for (i = 0; i < 192; i++)
				{
					mem = pVid;
					for (j= 0; j < n_lines; j++)
					{
					/*	*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01); */

						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = 0;

						pVid++;
						if (pVid >= Chipset.Video + 0x2000) pVid=Chipset.Video + 0x0800;
					}
					pLcd += hc;
					pVid = mem;
					for (j= 0; j < n_lines; j++)
					{
					/*	*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01); */

						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = 0;
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = 0;

						pVid++;
						if (pVid >= Chipset.Video + 0x2000) pVid=Chipset.Video + 0x0800;
					}
					pLcd += hc;
					pVid = mem;
					for (j= 0; j < n_lines; j++)
					{
					/*	*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x80)>>7;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x40)>>6;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x20)>>5;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x10)>>4;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x08)>>3;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x04)>>2;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x02)>>1;
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01);
						*pLcd++ = (*pVid & 0x01); */

						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 
						*pLcd++ = 0; 

						pVid++;
						if (pVid >= Chipset.Video + 0x2000) pVid=Chipset.Video + 0x0800;
					}
					pLcd += hc;
				}
			}
		}
	}
#endif

doit:

	EnterCriticalSection(&csGDILock);		// solving NT GDI problems (blit on windows display)
	{
		BitBlt(hWindowDC, nLcdX, nLcdY, nLcdCol[nLcdxZoom], nLcdRow[nLcdyZoom], hLcdDC, 0, 0, SRCCOPY);
		GdiFlush();
	}
	LeaveCriticalSection(&csGDILock);

#if defined HP85
nothing:
#endif
	;

}

//
// Display for writing a byte at video memory address d
//
VOID WriteToMainDisplay(BYTE a, DWORD d)
{
	INT  x0, xmin, xmax;
	INT  y0, ymin, ymax, dc;
	DWORD n_lines, mask;
	INT   hc;
	LPBYTE	pLcd;

	#if defined DEBUG_DISPLAY
	{
		TCHAR buffer[256];
		wsprintf(buffer,_T("%04X: Write Main Display %x\n"),Chipset.R[4]|(Chipset.R[5]<<8),d);
		OutputDebugString(buffer);
	}
	#endif

#if defined HP87
	if ((Chipset.crtsts & 0x80) == 0x00)						// alpha mode ?
	{
		xmin = xmax = 0;											// box to update on windows display
		ymin = ymax = 0;

		dc = 0;														// vertical offset
		if (nLcdxZoom == 1)											// zoom 1
			hc = ((Chipset.crtsts & 0x08) == 0x08) ? 10:15;				// height of a character (pagesize24 or 16)
		else														// zoom 2
			hc = ((Chipset.crtsts & 0x08) == 0x08) ? 20:30;				// height of a character (pagesize24 or 16)
		n_lines = ((Chipset.crtsts & 0x08) == 0x08) ? 24:16;
		mask = ((Chipset.crtsts & 0x40) == 0x40) ? 0x3FC0:0x10E0;	// address for wrap (alphall or alpha)

		if (d < Chipset.crtsad) d += mask;							// wrap ?
		d -= Chipset.crtsad;										// offset from screen start
		if ((d>= 0) && (d < n_lines * 80))							// visible ?
		{
			y0 = ((d / 80)*hc+dc);										// where on Y
			if (nLcdxZoom == 1)											// zoom 1
				x0 = (d % 80)*8;											// where on X
			else														// zoom 2 or 3
#if defined VRAM64K
				x0 = (d % 80) * 12 + 28;							// where on X
#else
				x0 = (d % 80) * 13 + 24;							// where on X
#endif
			xmin = xmax = x0;											// min box
			ymin = ymax = y0;

			if ((Chipset.crtsts & 0x20) == 0x20)						// inverse video mode ?
			{
				if (nLcdxZoom == 1)										
				{
					BitBlt(hLcdDC, x0, y0, 8, 10, hFontDC, (int)(a) << 3, 0, NOTSRCCOPY);
				}
				else
				{
#if defined VRAM64K
					BitBlt(hLcdDC, x0, y0, 12, 20, hFontDC, ((int)(a) *12), 0, NOTSRCCOPY);
#else
					BitBlt(hLcdDC, x0, y0, 13, 20, hFontDC, ((int)(a) << 4) + 3, 0, NOTSRCCOPY);
#endif
				}
			}
			else														// normal video mode
			{
				if (nLcdxZoom == 1)
				{
					BitBlt(hLcdDC, x0, y0, 8, 10, hFontDC, (int)(a) << 3, 0, SRCCOPY);
				}
				else
				{
#if defined VRAM64K
					BitBlt(hLcdDC, x0, y0, 12, 20, hFontDC, ((int)(a) *12), 0, SRCCOPY);
#else
					BitBlt(hLcdDC, x0, y0, 13, 20, hFontDC, ((int)(a) << 4) + 3, 0, SRCCOPY);
#endif
				}
			}
			if (nLcdxZoom == 1)
				xmax += 8;												// adjust max box
			else
				xmax += 13;
			ymax += hc;
		}
	}
	else														// graph mode
	{
		_ASSERT( d >= 0x10e0);

//#if !defined VRAM64K
		if (d < 0x10E0) d += 0x4000;							// wrap ?
//#endif
		d -= 0x10E0;											// offset from screen start
		if ((Chipset.crtsts & 0x40) == 0x00)					// graph mode ?
		{
#if defined VRAM64K
			n_lines = 127;
			hc = 0;
#else
			n_lines = 50;											// 50 bytes per line (50 * 8 = 400)
			if (nLcdxZoom == 1)										// zoom 1 ?
				hc = 120;												// offset (640 - 400) / 2 
			else													// zoom 2
				hc = 144;												// offset (1088 - 800) / 2
#endif
		}
		else													// graphall mode
		{
#if defined VRAM64K
			n_lines = 127;
			hc = 0;
#else
			n_lines = 68;											// 68 bytes per line (68 * 8 = 544)
			if (nLcdxZoom == 1)										// zoom 1 ?
				hc = 48;												// offset (640 - 544) / 2
			else													// zoom 2
				hc = 0;													// offset (1088 - 1088) / 2
#endif
		}
#if defined VRAM64K
		y0 = (d / n_lines);
#else
		y0 = (d / n_lines) * nLcdyZoom;							// Y location
#endif
		x0 = (d % n_lines);										// X location

		if (nLcdxZoom == 1)										// zoom 1 ?
			pLcd = pbyLcd + hc + x0*8 + y0*640;						// address on screen bitmap
		else													// zoom 2
#if defined VRAM64K
			pLcd = pbyLcd + hc + x0*8 + y0*1016;
#else
			pLcd = pbyLcd + hc + x0*16 + y0*1088;					// address on screen bitmap
#endif

		if ((Chipset.crtsts & 0x20) == 0x20)					// inverse video mode ?
		{
			if (nLcdxZoom == 1)										// zoom 1 ?
			{
				*pLcd++ = 1 - ((a & 0x80)>>7);							// first pixel
				*pLcd++ = 1 - ((a & 0x40)>>6);							// second pixel
				*pLcd++ = 1 - ((a & 0x20)>>5);							// ...
				*pLcd++ = 1 - ((a & 0x10)>>4);
				*pLcd++ = 1 - ((a & 0x08)>>3);
				*pLcd++ = 1 - ((a & 0x04)>>2);
				*pLcd++ = 1 - ((a & 0x02)>>1);
				*pLcd++ = 1 - (a & 0x01);
			}
			else													// zoom 2
			{
				*pLcd++ = 1 - ((a & 0x80)>>7);							// first pixel first time
#if !defined VRAM64K
				*pLcd++ = 1 - ((a & 0x80)>>7);							// first pixel second time
#endif
				*pLcd++ = 1 - ((a & 0x40)>>6);							// second pixel first time
#if !defined VRAM64K
				*pLcd++ = 1 - ((a & 0x40)>>6);							// ...
#endif
				*pLcd++ = 1 - ((a & 0x20)>>5);
#if !defined VRAM64K
				*pLcd++ = 1 - ((a & 0x20)>>5);
#endif
				*pLcd++ = 1 - ((a & 0x10)>>4);
#if !defined VRAM64K
				*pLcd++ = 1 - ((a & 0x10)>>4);
#endif
				*pLcd++ = 1 - ((a & 0x08)>>3);
#if !defined VRAM64K
				*pLcd++ = 1 - ((a & 0x08)>>3);
#endif
				*pLcd++ = 1 - ((a & 0x04)>>2);
#if !defined VRAM64K
				*pLcd++ = 1 - ((a & 0x04)>>2);
#endif
				*pLcd++ = 1 - ((a & 0x02)>>1);
#if !defined VRAM64K
				*pLcd++ = 1 - ((a & 0x02)>>1);
#endif
				*pLcd++ = 1 - (a & 0x01);
#if !defined VRAM64K
				*pLcd++ = 1 - (a & 0x01);
#endif
#if !defined VRAM64K
				pLcd += 1088 - 16;										// next screen line
				*pLcd++ = 1 - ((a & 0x80)>>7);							// first pixel ...
				*pLcd++ = 1 - ((a & 0x80)>>7);
				*pLcd++ = 1 - ((a & 0x40)>>6);
				*pLcd++ = 1 - ((a & 0x40)>>6);
				*pLcd++ = 1 - ((a & 0x20)>>5);
				*pLcd++ = 1 - ((a & 0x20)>>5);
				*pLcd++ = 1 - ((a & 0x10)>>4);
				*pLcd++ = 1 - ((a & 0x10)>>4);
				*pLcd++ = 1 - ((a & 0x08)>>3);
				*pLcd++ = 1 - ((a & 0x08)>>3);
				*pLcd++ = 1 - ((a & 0x04)>>2);
				*pLcd++ = 1 - ((a & 0x04)>>2);
				*pLcd++ = 1 - ((a & 0x02)>>1);
				*pLcd++ = 1 - ((a & 0x02)>>1);
				*pLcd++ = 1 - (a & 0x01);
				*pLcd++ = 1 - (a & 0x01);
#endif
			}
		}
		else													// normal video mode
		{
			if (nLcdxZoom == 1)										// zoom 1 ?
			{
				*pLcd++ = (a & 0x80)>>7;
				*pLcd++ = (a & 0x40)>>6;
				*pLcd++ = (a & 0x20)>>5;
				*pLcd++ = (a & 0x10)>>4;
				*pLcd++ = (a & 0x08)>>3;
				*pLcd++ = (a & 0x04)>>2;
				*pLcd++ = (a & 0x02)>>1;
				*pLcd++ = (a & 0x01);
			}
			else													// zoom 2
			{
				*pLcd++ = (a & 0x80)>>7;
#if !defined VRAM64K
				*pLcd++ = (a & 0x80)>>7;
#endif
				*pLcd++ = (a & 0x40)>>6;
#if !defined VRAM64K
				*pLcd++ = (a & 0x40)>>6;
#endif
				*pLcd++ = (a & 0x20)>>5;
#if !defined VRAM64K
				*pLcd++ = (a & 0x20)>>5;
#endif
				*pLcd++ = (a & 0x10)>>4;
#if !defined VRAM64K
				*pLcd++ = (a & 0x10)>>4;
#endif
				*pLcd++ = (a & 0x08)>>3;
#if !defined VRAM64K
				*pLcd++ = (a & 0x08)>>3;
#endif
				*pLcd++ = (a & 0x04)>>2;
#if !defined VRAM64K
				*pLcd++ = (a & 0x04)>>2;
#endif
				*pLcd++ = (a & 0x02)>>1;
#if !defined VRAM64K
				*pLcd++ = (a & 0x02)>>1;
#endif
				*pLcd++ = (a & 0x01);
#if !defined VRAM64K
				*pLcd++ = (a & 0x01);
#endif
#if !defined VRAM64K
				pLcd += 1088 - 16;
				*pLcd++ = (a & 0x80)>>7;
				*pLcd++ = (a & 0x80)>>7;
				*pLcd++ = (a & 0x40)>>6;
				*pLcd++ = (a & 0x40)>>6;
				*pLcd++ = (a & 0x20)>>5;
				*pLcd++ = (a & 0x20)>>5;
				*pLcd++ = (a & 0x10)>>4;
				*pLcd++ = (a & 0x10)>>4;
				*pLcd++ = (a & 0x08)>>3;
				*pLcd++ = (a & 0x08)>>3;
				*pLcd++ = (a & 0x04)>>2;
				*pLcd++ = (a & 0x04)>>2;
				*pLcd++ = (a & 0x02)>>1;
				*pLcd++ = (a & 0x02)>>1;
				*pLcd++ = (a & 0x01);
				*pLcd++ = (a & 0x01);
#endif
			}
		}
		if (nLcdxZoom == 1)										// update min box and max box
		{
			xmin = x0*8 + hc;
			ymin = y0;
			xmax = xmin + 8;
			ymax = ymin + 1;
		}
		else
		{
#if defined VRAM64K
			xmin = x0*8 + hc;
			ymin = y0;
			xmax = xmin + 8;
			ymax = ymin + 1;
#else
			xmin = x0*16 + hc;
			ymin = y0;
			xmax = xmin + 16;
			ymax = ymin + 2;
#endif
		}
	}
#endif

#if defined HP85
	if ((Chipset.crtwsts & 0x80) == 0x00)		// alpha mode	
	{
		xmin = xmax = 0;
		ymin = ymax = 0;

		if (nLcdxZoom == 1)
		{
			hc = 12;
			dc = 0;
		}
		else if (nLcdxZoom == 2)
		{
			hc = 24;
			dc = 0;
		}
		else if (nLcdxZoom == 3)
		{
			hc = 36;
			dc = 0;
		}
		n_lines = 16;
		mask = 0x800;	// 64 lines


		if (d < (Chipset.crtsad >> 1)) d += mask;					// wrap
		d -= (Chipset.crtsad >> 1);									// offset from screen start
		if ((d>= 0) && (d < n_lines * 32))							// visible ?
		{
			y0 = ((d / 32)*hc+dc);
			if (nLcdxZoom == 1)
				x0 = (d % 32)*8;
			else if (nLcdxZoom == 2)
				x0 = (d % 32)*16;
			else if (nLcdxZoom == 3)
				x0 = (d % 32)*24;
			xmin = xmax = x0;
			ymin = ymax = y0;

			if (nLcdxZoom == 1)
			{
				BitBlt(hLcdDC, x0, y0, 8, 12, hFontDC, (int)(a) << 3, 0, SRCCOPY);
			}
			else if (nLcdxZoom == 2)
			{
				BitBlt(hLcdDC, x0, y0, 16, 24, hFontDC, (int)(a) << 4, 0, SRCCOPY);
			}
			else if (nLcdxZoom == 3)
			{
				BitBlt(hLcdDC, x0, y0, 24, 36, hFontDC, (int)(a) * 24, 0, SRCCOPY);
			}
		}
		else
			return;

		if (nLcdxZoom == 1)
			xmax += 8;									// adjust rectangle size
		else if (nLcdxZoom == 2)
			xmax += 16;
		else if (nLcdxZoom == 3)
			xmax += 24;
		ymax += hc;
	}
	else
	{
//		if (d < 0x800) d += 0x800;
		if ((d < (Chipset.crtsad >> 1)) || (d >= (Chipset.crtsad >> 1) + 0x3000))
			return;
		d -= (Chipset.crtsad >> 1);

		n_lines = 32;		// 256 pixels / line
		hc = 0;
		y0 = (d / n_lines) * nLcdyZoom;
		x0 = (d % n_lines);

		if (nLcdxZoom == 1)
			pLcd = pbyLcd + hc + x0*8 + y0*256;
		else if (nLcdxZoom == 2)
			pLcd = pbyLcd + hc + x0*16 + y0*512;
		else if (nLcdxZoom == 3)
			pLcd = pbyLcd + hc + x0*24 + y0*768;

		if (nLcdxZoom == 1)
		{
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x01);
		}
		else if (nLcdxZoom == 2)
		{
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x01)>>0;
			*pLcd++ = (a & 0x01)>>0;
			pLcd += 512 - 16;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x01)>>0;
			*pLcd++ = (a & 0x01)>>0;
		}
		else if (nLcdxZoom == 3)
		{
		/*	*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x01);
			*pLcd++ = (a & 0x01);
			*pLcd++ = (a & 0x01);
			pLcd += 768 - 24;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x01);
			*pLcd++ = (a & 0x01);
			*pLcd++ = (a & 0x01);
			pLcd += 768 - 24;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x01);
			*pLcd++ = (a & 0x01);
			*pLcd++ = (a & 0x01); */

			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x01);
			*pLcd++ = (a & 0x01);
			*pLcd++ = 0;
			pLcd += 768 - 24;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = (a & 0x80)>>7;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = (a & 0x40)>>6;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = (a & 0x20)>>5;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = (a & 0x10)>>4;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = (a & 0x08)>>3;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = (a & 0x04)>>2;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = (a & 0x02)>>1;
			*pLcd++ = 0;
			*pLcd++ = (a & 0x01);
			*pLcd++ = (a & 0x01);
			*pLcd++ = 0;
			pLcd += 768 - 24;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
			*pLcd++ = 0;
		}
		if (nLcdxZoom == 1)
		{
			xmin = x0*8 + hc;
			ymin = y0;
			xmax = xmin + 8;
			ymax = ymin + 1;
		}
		else if (nLcdxZoom == 2)
		{
			xmin = x0*16 + hc;
			ymin = y0;
			xmax = xmin + 16;
			ymax = ymin + 2;
		}
		else if (nLcdxZoom == 3)
		{
			xmin = x0*24 + hc;
			ymin = y0;
			xmax = xmin + 24;
			ymax = ymin + 3;
		}
	}
#endif

	EnterCriticalSection(&csGDILock);						// blit the min-max box on window display
	{
		BitBlt(hWindowDC, nLcdX+xmin, nLcdY+ymin, xmax-xmin, ymax-ymin, hLcdDC, xmin, ymin, SRCCOPY);
		GdiFlush();
	}
	LeaveCriticalSection(&csGDILock);
}

//
// update display annuciators (via kml.c)
//
VOID UpdateAnnunciators(VOID)
{
	int i;
	DWORD j=1;

#if defined HP87
	for (i = 1; i < 22; ++i)
#endif
#if defined HP85
	for (i = 1; i < 11; ++i)
#endif
	{
		if (Chipset.annun & j)
			DrawAnnunciator(i, TRUE);
		else
			DrawAnnunciator(i, FALSE);
		j <<= 1;
	}
}

//
// Main window resize
//
VOID ResizeWindow(VOID)
{
	RECT rectWindow;
	RECT rectClient;

	if (hWnd == NULL) return;				// return if window closed

	rectWindow.left   = 0;
	rectWindow.top    = 0;
	rectWindow.right  = nBackgroundW;
	rectWindow.bottom = nBackgroundH;
	AdjustWindowRect(&rectWindow, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);
	SetWindowPos (hWnd, (HWND)NULL, 0, 0,
		rectWindow.right  - rectWindow.left,
		rectWindow.bottom - rectWindow.top,
		SWP_NOMOVE | SWP_NOZORDER);
	GetClientRect(hWnd, &rectClient);
	AdjustWindowRect(&rectClient, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);
	if (rectClient.bottom < rectWindow.bottom)
	{
		rectWindow.bottom += (rectWindow.bottom - rectClient.bottom);
		SetWindowPos (hWnd, (HWND)NULL, 0, 0,
			rectWindow.right  - rectWindow.left,
			rectWindow.bottom - rectWindow.top,
			SWP_NOMOVE | SWP_NOZORDER);
	}

	_ASSERT(hWindowDC);						// move destination window
	SetWindowOrgEx(hWindowDC, nBackgroundX, nBackgroundY, NULL);
	InvalidateRect(hWnd,NULL,TRUE);
}

