/*
 *   Iop.c
 *
 *   Copyright 2004 Olivier De Smet
 */

//
//   IO Processors simulation based on behaviors from
//
//   "Assembly language I/O reference manual" HP-83/85 Number 00085-90818
//   ASM_IO_REF_83_85.pdf from Andr Koppel
//   "Original 320 rom listing from HP"
//   ROM320_P1.pdf from Andr Koppel
// 

#include "StdAfx.h"
#include "HP8x.h"
#include "io.h"								// I/O definitions

//#define DEBUG_IOP

#if defined DEBUG_IOP
static WORD IOP_ADR[] = {					// states of automata
	10000, 10100,
	10200, 10300,
	10400, 10500,
	10600, 10700,
	10800, 10900,
	0, 201, 301
};

static TCHAR *IOP_TXT[] ={					// labels of states of automata 
	_T("Read Status"), _T("Input"),
	_T("Burst"), _T("Interrupt control"),
	_T("Interface control"), _T("Read Auxilliary"),
	_T("Write Control"), _T("Output"),
	_T("Send Commands"), _T("Write Auxilliary"),
	_T("idle"), _T("got interrupt"), _T("got command")
};

#define NLAB (sizeof(IOP_ADR)/sizeof(IOP_ADR[0]))

static TCHAR *IOP_LAB(WORD d)				// find a label for the state d
{
	WORD i;

	for(i = 0; i < NLAB; i++)
		if (IOP_ADR[i] == d)
			return IOP_TXT[i];
	return NULL;
}
#endif

//################
//#
//#    Public functions
//#
//################

//
// main Iop simualtion loop
//   FSA implementation, some hacks for Auxilliary registers (no docs)
//   used in MS roms to preserve the EOL state between Discs operations
//   (see pdf of MS rom)
//   

__inline VOID __fastcall IopCmd(IOP *i)
{
	#if defined DEBUG_IOP
	WORD iopst;
	#endif

	#if defined DEBUG_IOP
	iopst = i->iopst;
	#endif

	if (i->ccrwr)
	{
		if ((i->ccr & 0x80) == 0x80)	// reset iop
		{
			i->ccrwr = FALSE;
			i->iopst = 100;
		}

		else if ((i->ccr & 0x01) == 0x01)	// int iop
		{
			i->ccrwr = FALSE;
			i->iopst = 200;
		}

		else if ((i->ccr & 0x02) == 0x02)	// com iop
		{
			i->ccrwr = FALSE;
			i->psr |= 0x02;						// set busy in psr
			i->iopst = 300;
		}
	}

	switch(i->iopst)
	{
		BYTE b, st;

	// WAIT
		case 0:		
			i->psr &= 0xFD;					// busy = 0 in psr
			i->psr &= 0xFB;					// ped = 0 in psr
			i->iopst++;
		case 1:										// idle
			break;
	// RESET
		case 100:									// reset step 0	
			ZeroMemory(i->streg, 16);
			ZeroMemory(i->ctreg, 32);
			ZeroMemory(i->axreg, 16);
			((BOOL (*)(IOP *))i->iop_reset)(i);	// reset the interface
			i->ioptimer = 614;				// wait 1ms
			i->iopst++;
			break;
		case 101:
			if (i->ioptimer == 0)			// wait done
				i->iopst++;
			break;
		case 102:
			i->intiop = TRUE;				// set interrupt
			i->ib = 0x55;					// first scratch data
			i->psr |= 0x01;					// set IBF in psr
			i->iopst++;
			break;
		case 103:
			if (Chipset.inte & (bInterrupt == FALSE))
			{
				bInterrupt = TRUE;				// take interrupt
				i->iopst++;
			}
			break;
		case 104:
			if ((i->psr & 0x01) == 0x00)	// interrupt taken by CPU (first IB read done)
				i->iopst++;
			break;
		case 105:
			i->ib = 0x3;					// self-test passed data
			i->psr |= 0x01;					// set IBF in psr
			i->iopst++;
			break;
		case 106:
			if ((i->psr & 0x01) == 0x00)	// second IB read done
				i->iopst = 0;
			break;
	// INT
		case 200:
			if ((i->psr & 0x08) == 0x08)
				i->psr &= 0xF7;				// clear PACK in psr
			else
				i->psr |= 0x08;				// set PACK in psr
			i->iopst++;		
			break;
		case 201:
			i->iopst = 0;
			break;
	// COM
		case 300:
			if ((i->psr & 0x80) == 0x80)	// OBF set in psr ?
				i->iopst++;
			break;
		case 301:
			i->psr &= 0x7F;					// clear OBF in psr
			switch (i->ob >> 4)
			{
				case 0:				// Read status register		0000
					i->dt = i->ob & 0x0F;					// # register
					i->iopst = 10000;
					break;
				case 1:				// Input					0001
					i->dt = i->ob & 0x0F;					// kind of input    bit 0: IOP term (EOI)
															//                  bit 1: interrupt input 
															//                  bit 2: char term (control reg 27)
															//					bit 3: count term (control reg 25,26)
					_ASSERT((i->dt == 0x00) || (i->dt == 0x01) || (i->dt == 0x04));
					i->iopst = 10100;
					break;
				case 2:				// Burst					0010
					i->psr &= 0xFD;									// busy = 0 in psr
					i->dt = i->ob & 0x0F;					// direction & termination
					_ASSERT(i->dt < 2);
					if (i->dt == 0x01)
						i->iopst = 10200;							// burst in with EOI
					if (i->dt == 0x00)								
						i->iopst = 10250;							// burst out
					break;
				case 3:				// Interrupt control		0011
					i->iopst = 10300;
					break;
				case 4:				// Interface control		0100
					i->dt = i->ob & 0x0F;					// command
					i->iopst = 10400;
					break;
				case 5:				// unused					0101
				case 6:				// unused					0110
					_ASSERT(0);
					break;
				case 7:				// Read Auxilliary			0111		for HP-IB
					i->dt = i->ob & 0x0F;					// # register
					i->iopst = 10500;
					break;
				case 8:				// Write control 0			1000
				case 9:				// Write control 1			1001
					i->dt = i->ob & 0x1F;					// # register
					i->iopst = 10600;
					break;
				case 10:			// Output					1010
					i->psr &= 0xFD;					// busy = 0 in psr
					i->iopst = 10700;
					break;
				case 11:			// Send Commands			1011		emit data with ATN=1 over HP-IB (command) 
					i->psr &= 0xFD;									// busy = 0 in psr
					i->iopst = 10800;
					break;
				case 12:			// unused					1100
				case 13:			// unused					1101
					_ASSERT(0);
					break;
				case 14:			// Write Auxilliary			1110		for HP-IB
					i->dt = i->ob & 0x0F;					// # register
					i->iopst = 10900;
					break;
				case 15:			// Extension				1111
					_ASSERT(0);
					i->iopst = 11000;
					break;
			}
			break;
	// Read status register
		case 10000:
			i->ib = i->streg[i->dt];		// put status reg in IB
			i->psr |= 0x01;									// set IBF in psr
			i->dt++;
			i->dt &= 0x0F;
			i->iopst++;
			break;
		case 10001:
			if ((i->psr & 0x80) == 0x80)					// OBF set in psr ?
			{
				i->psr &= 0x7F;									// clear OBF in psr
				i->iopst--;
			}
			if ((i->ccr & 0x04) == 0x04)					// CED set in ccr ?
				i->iopst = 0;
			break;
	// Input
		case 10100:
			if ((i->ccr & 0x04) == 0x00)					// CED clear in ccr ?
				i->iopst++;
			break;
		case 10101:	
			if ((i->ccr & 0x04) == 0x04)					// CED set in ccr ?
				i->iopst = 0;
			if (((BOOL (*)(BYTE *, BYTE *))i->iop_input)(&b, &st))
			{
				i->ib = b;									
				if ((st & 0x80) == 0x80)								// end of input ?
				{
					i->psr |= 0x04;								// set PED in psr
					i->iopst = 10103;
				}
				i->psr |= 0x01;									// set IBF in psr
				i->iopst++;
			}
			break;
		case 10102:
			if ((i->psr & 0x01) == 0x00)					// IBF clear in psr ?
				i->iopst++;
			break;
		case 10103:
			if ((i->psr & 0x80) == 0x80)					// OBF set in psr ?
			{	
				i->psr &= 0x7F;								// clear OBF in psr
				i->iopst = 10101;
			}
			break;
		case 10104:
			if ((i->psr & 0x01) == 0x00)					// IBF clear in psr ?
				i->iopst = 0;
			break;
	// Burst in 256 bytes
		case 10200:
			i->dt = 0;										// set counter
			i->iopst++;
			break;
		case 10201:
			if ((i->psr & 0x80) != 0x80)					// OBF not set in psr ?
				break;
			i->psr &= 0x7F;									// clear OBF in psr
			i->iopst++;
		case 10202:		// get a byte from outside to send it to cpu
			if (((BOOL (*)(BYTE *, BYTE *))i->iop_input)(&b, &st))
			{
				i->ib = b;
				i->psr |= 0x01;									// set IBF in psr
				i->dt++;										// one byte done
				i->iopst++;
				break;
			}
			_ASSERT(0);
			break;
		case 10203:
			if ((i->psr & 0x01) == 0x00)					// IBF clear in psr ?
			{
				if (i->dt != 0)
					i->iopst--;									// next byte
				else
				{
					i->ib = 0x55;								// dummy byte
					i->psr |= 0x01;								// set IBF in psr
					i->iopst++;
				}
			}
			break;
		case 10204:
			if ((i->psr & 0x01) == 0x00)					// IBF clear in psr ?
			{
				i->intiop = TRUE;								// set interrupt
				i->ib = 0xAA;									// first scratch data
				i->psr |= 0x01;									// set IBF in psr
				_ASSERT(Chipset.inte & (bInterrupt == FALSE));
				bInterrupt = TRUE;
				i->iopst++;
			}
			break;
		case 10205:
			if ((i->psr & 0x01) == 0x00)					// interrupt taken by CPU (first IB read done)
				i->iopst++;
			break;
		case 10206:
			i->ib = 0x1;									// burst interrupt
			i->psr |= 0x01;									// set IBF in psr
			i->iopst++;
			break;
		case 10207:
			if ((i->psr & 0x01) == 0x00)					// second IB read done
				i->iopst = 0;
			break;
	// Burst out 256 bytes
		case 10250:	
			i->dt = 0;										// set counter
			i->iopst++;
			break;
		case 10251:		// get a byte from cpu to send it outside
			if ((i->psr & 0x80) != 0x80)					// OBF not set in psr ?
				break;
			i->psr &= 0x7F;									// clear OBF in psr
			i->dt++;										// one more
			if (i->dt != 0)									
			{
				((BOOL (*)(BYTE, BYTE))i->iop_output)(i->ob, 0);
				break;
			}
			else													// last one
			{
				((BOOL (*)(BYTE, BYTE))i->iop_output)(i->ob, 0x80);
				i->intiop = TRUE;								// set interrupt
				i->ib = 0xAA;									// first scratch data
				i->psr |= 0x01;									// set IBF in psr
				_ASSERT(Chipset.inte & (bInterrupt == FALSE));
				bInterrupt = TRUE;
				i->iopst++;
			}
			break;
		case 10252:
			if ((i->psr & 0x01) == 0x00)					// interrupt taken by CPU (first IB read done)
				i->iopst++;
			break;
		case 10253:
			i->ib = 0x1;									// burst interrupt
			i->psr |= 0x01;									// set IBF in psr
			i->iopst++;
			break;
		case 10254:
			if ((i->psr & 0x01) == 0x00)					// second IB read done
				i->iopst = 0;
			break;
	// Interupt control
		case 10300:													// do nothing
			break;
	// Interface command
		case 10400:
			if (i->dt <= 9)
				((VOID (*)(IOP *))i->iop_ifc[i->dt])(i);
			else
				_ASSERT(0);
			i->iopst = 0;
			break;
	// Read Auxilliary register (only used for HPIB actually, so some assert)
		case 10500:
			_ASSERT(i->dt == 0x07 && i->iop == 1);
			switch (i->axreg[6])
			{
				case 0x38:
					i->ib = i->ctreg[16];		// put ct reg 16 in IB
					break;
				default:
					_ASSERT(0);
					break;
			}
			i->psr |= 0x01;								// set IBF in psr
			i->dt++;
			i->dt &= 0x0F;
			i->iopst++;
			break;
		case 10501:
			if ((i->psr & 0x80) == 0x80)				// OBF set in psr ?
			{
				i->psr &= 0x7F;							// clear OBF in psr
				i->iopst--;
			}
			if ((i->ccr & 0x04) == 0x04)				// CED set in ccr ?
				i->iopst = 0;
			break;
	// Write control
		case 10600:
			if ((i->psr & 0x80) == 0x80)				// OBF set in psr ?
			{
				i->psr &= 0x7F;								// clear OBF in psr
				i->ctreg[i->dt] = i->ob;	// put OB in control reg
				i->dt++;
				i->dt &= 0x1F;
				if ((i->ccr & 0x04) == 0x04)				// CED set in ccr ?
					i->iopst = 0;
			}
			break;
	// Output
		case 10700:
			if ((i->ccr & 0x04) == 0x00)				// CED clear in ccr ?
				i->iopst++;
			break;
		case 10701:
			if ((i->psr & 0x80) == 0x80)				// OBF set in psr ?
			{
				st = 0x00;
				if ((i->ccr & 0x04) == 0x04)
					 st = 0x80;
				if (((BOOL (*)(BYTE, BYTE))i->iop_output)(i->ob, st))
				{
					i->psr &= 0x7F;								// clear OBF in psr
					if ((i->ccr & 0x04) == 0x04)				// CED set in ccr ?
						i->iopst++;
				}
			}
			break;
		case 10702:
			((VOID (*)(IOP *))i->iop_ifc[7])(i);		// send EOL sequence
			i->iopst = 0;
			break;
	// Send Commands
		case 10800:
			if ((i->psr & 0x80) == 0x80)				// OBF set in psr ?
			{
				i->psr &= 0x7F;								// clear OBF in psr
				((BOOL (*)(BYTE))i->iop_send)(i->ob);					// send as command over HPIB
				i->dt++;
				i->dt &= 0x1F;
				if ((i->ccr & 0x04) == 0x04)				// CED set in ccr ?
					i->iopst = 0;
			}
			break;
	// Write Auxilliary (only used for HPIB actually, so some assert)
		case 10900:
			_ASSERT(i->dt == 0x06 && i->iop == 1);
			if ((i->psr & 0x80) == 0x80)				// OBF set in psr ?
			{
				
				i->psr &= 0x7F;								// clear OBF in psr
				i->axreg[i->dt] = i->ob;	// put OB in auxilliary reg
				i->dt++;
				i->dt &= 0x0F;
				if ((i->ccr & 0x04) == 0x04)				// CED set in ccr ?
					i->iopst = 0;
			}
			break;
	// NNN
		default:
			_ASSERT(0);
	}

	#if defined DEBUG_IOP
	{
		TCHAR buffer[256];
		TCHAR *TXT = IOP_LAB((WORD)(i->iopst));
		int k;
		unsigned long l;

		if (iopst != i->iopst)
		{
			if (TXT == NULL)
				k = wsprintf(buffer,_T("%04X: IOP : %u:%u->%u\n"), Chipset.R[4]|(Chipset.R[5]<<8), i, iopst,i->iopst);
			else
				k = wsprintf(buffer,_T("%04X: IOP : %u:%u->%u ( %s )\n"), Chipset.R[4]|(Chipset.R[5]<<8), i, iopst,i->iopst, TXT);
#if defined DEBUG_ONDISK
			WriteFile(hDebug, buffer, (unsigned long) (k), &l, NULL); 
//			FlushFileBuffers(hDebug);
#endif
			OutputDebugString(buffer);
		}
	}
	#endif

}

//
// reload function for iop's and perform initialization on peripherals
//
VOID IopReload(VOID)
{
	if (Chipset.iop[7].iop == 1)				// iop S.C. 7 present as HPIB controller (see hpib.c)
												//			HP9121 floppy disk ADDR 0
												//			HP2225 printer ADDR 1
												//			HP7470 tracer ADDR 5
	{
		Chipset.iop[7].iop_ifc = hpib_ifc;
		Chipset.iop[7].iop_input = hpib_receive;		// (BOOL)(BYTE *data, BYTE *EOI)
		Chipset.iop[7].iop_output = hpib_send_d;		// (BOOL)(BYTE data, BYTE EOI)
		Chipset.iop[7].iop_send = hpib_send_c;			// (BOOL)(BYTE)
		Chipset.iop[7].iop_reset = hpib_reset;			// (BOOL)(int)
		hpib_reset(&Chipset.iop[7]);
	}

	if (Chipset.iop[10].iop == 2)				// iop S.C. 10 present as serial controller HP82939A (see serial.c)
	{
		Chipset.iop[10].iop_ifc = serial_ifc;
		Chipset.iop[10].iop_input = serial_receive;	// (BOOL)(BYTE *data, BYTE *EOI)
		Chipset.iop[10].iop_output = serial_send_d;	// (BOOL)(BYTE data, BYTE EOI)
		Chipset.iop[10].iop_send = serial_send_c;	// (BOOL)(BYTE)
		Chipset.iop[10].iop_reset = serial_reset;	// (BOOL)(int)
		serial_reset(&Chipset.iop[10]);
	}

#if defined HP85
	Prt85_init();
#endif
}

//
// stop iop's and free memory
//
VOID IopStop(VOID)
{
	if (Chipset.iop[7].iop == 1)
		hpib_stop(&Chipset.iop[7]);
	if (Chipset.iop[10].iop == 2)
		serial_stop(&Chipset.iop[10]);
#if defined HP85
	Prt85_close();
#endif
}

